clear variables
close all
clc

%% Print Information about Setting  
L = 1; % length of the string
c = 0.10; % wave velocity
fprintf('We consider the control of the vertical deflection of a string of length %f.\n', L);
fprintf('The initial deflection is: y(x,0) = 3sin(pi*x/L).\n')
fprintf('We set Neumann b.c.s on both sides of the string: d/dx y(0,t) = 0, d/dx y(1,t) = 0.\n');

%% Space and Time Grid (Set)
Nx = 11; % total number of spatial nodes
dx = L/(Nx-1); % step in position

T_max = 50; % T_max
fprintf('The total simulation time is T_max = %.2f sec.\n', T_max);
fprintf('\n')
N_tot = 1001; % total number of temporal nodes 
tgrid = linspace(0,T_max,N_tot); 
dt = tgrid(2) - tgrid(1); % time step

fprintf('L: %.2f m,\t T_max: %.2f sec\n', L, T_max);
fprintf('dx: %.2f m,\t dt: %.2f sec\n', dx, dt);
fprintf('\n')

%% Prediction Horizon (Set)
Nt = 200; % total number of temporal nodes within one prediction horizon
fprintf('Nt: %d ,\t T: %.2f sec\n', Nt, dt*Nt);
fprintf('\n')

%% Value of tau (Input)
prompt = 'Choose a multiple of dt, which serves as a value for tau.\n';
tau = input(prompt);
if mod(tau/dt, 1) ~= 0
    fprintf('Please choose a different value, such that tau/dt is an integer. dt is set to %f sec. \n', dt)
    prompt = 'Choose a new multiple of dt / a new value of tau.\n';
    tau = input(prompt);    
end
N_MPC = tau;
tau = tau*dt;
fprintf('\n')
fprintf('tau set to: %.3f sec.\n', tau); 
fprintf('T-tau: %.3f sec.\n', dt*Nt-tau); % has to be positive
fprintf('\n')

%% Value of h (Input)
prompt = 'Choose a multiple of dt, which serves as a value for h.\n';
h = input(prompt);
if mod(Nt/h, 1) ~= 0
    fprintf('Please choose a different value, such that Nt/h is an integer. Nt is set to: %f.\n', Nt)
    prompt = 'Choose a new multiple of dt / a new value of h.\n';
    h = input(prompt);    
end
fprintf('\n')
fprintf('h set to: %.3f*dt \n', h); % has to be positive
fprintf('\n')
K = Nt/h; % must be an integer
h = h*dt; 

%% Set Matrices
B = zeros(2*Nx,1);
B(end) = 1; % control acts at the right end
Q = 30*eye(2*Nx,2*Nx);
R = 1;

%% Reference Solution: Load it or Compute it 

file_ref = ['ref_output.mat'];

if isfile(file_ref)
    % File exists.
    fprintf('\n')
    fprintf('Loading reference solution...\n');
    fprintf('\n')
    load(file_ref);
else
    % File does not exist.
    fprintf('\n')
    fprintf('Starting to compute the reference solution...\n');
    fprintf('\n')
    [y_ref,u_ref] = state_reference(Nx,dx,N_tot,dt,c,L,B,Q,R);
    save(file_ref,'y_ref','u_ref');
end


%% RBM-MPC Solution: Load it or Compute it 

file_RBM_MPC = ['tau_' num2str(tau, '%.4f')  '_h_' num2str(h, '%.4f') '.mat'];

if isfile(file_RBM_MPC)
    % File exists.
    fprintf('\n')
    fprintf('Loading RBM-MPC solution...\n');
    fprintf('\n')
    load(file_RBM_MPC);
else
    % File does not exist.
    fprintf('\n')
    fprintf('Starting to compute the RBM-MPC solution...\n');
    fprintf('\n')
    [y_tot, u_tot] = state_RBM_MPC(Nx,dx,N_tot,Nt,N_MPC,K,dt,c,L,B,Q,R);
    fprintf('RBM_MPC solution computed!\n');
    fprintf('\n')
    save(file_RBM_MPC,'y_tot','u_tot');
end

%% Plot Applied Control
figure(1)
hold on
title('Applied Control')
xlabel 't [s]'
ylabel 'u(t)'
t = linspace(0+dt,T_max-dt,N_tot-1).';
plot(t,u_ref,'DisplayName','Reference')
plot(t,u_tot,'DisplayName','RBM-MPC')
legend('Reference','RBM-MPC')
ylim([-13 13])
plot_name = ['Control_' num2str(tau, '%.3f') '_' num2str(h,'%.3f') '.png'];
saveas(gcf,plot_name);

%% Plot Trajectory

create_movie_d = 1;  % do you want to create a movie with the deflection?  Set to zero if not. 
if create_movie_d
     movie_name = ['Movie_Deflection_' num2str(tau, '%.3f') '_' num2str(h, '%.3f')];  % file name
     vidfile_deflection = VideoWriter(movie_name,'Uncompressed AVI');
     vidfile_deflection.FrameRate = 10;  % change this number to slow down or speed up the movie
     open(vidfile_deflection);
     fig = figure;
     set(fig,'color','w');
end

figure(2)
hold on
x = linspace(0,L,Nx).';
for i = 1:N_tot
    plot(x,y_ref(1:Nx,i),'DisplayName','Reference','LineWidth',3)
    hold on
    plot(x,y_tot(1:Nx,i),'DisplayName','RBM-MPC','LineWidth',3)
    ylim([-3.5 3.5])
    val = i*dt;
    title('Deflection at t = ', num2str(val,'%.2f'))
    xlabel 'x [m]'
    ylabel 'y(t) [m]'
    legend('Reference','RBM-MPC')

   if create_movie_d
       frame = getframe(fig);
       writeVideo(vidfile_deflection, frame);
   end

   pause(0.01);
   hold off;
end

if create_movie_d
    close(vidfile_deflection);
end