function [y_tot,u_tot] = state_RBM_MPC(Nx,dx,N_tot,Nt,N_MPC,K,dt,c,L,B,Q,R)
    %% FD Discretization of Laplacian
    A_0 = zeros(Nx,Nx);
    for ii = 2:Nx-1      
        A_0(ii,ii)   = -2*c.^2/dx.^2;
        A_0(ii,ii+1) =    c.^2/dx.^2;
        A_0(ii,ii-1) =    c.^2/dx.^2;
    end
    A_0(1,1) = -2*c.^2/dx.^2;
    A_0(1,2) = 2*c.^2/dx.^2; 
    A_0(Nx,Nx-1) =  2*c.^2/dx.^2; 
    A_0(Nx,Nx)   = -2*c.^2/dx.^2;
    
    %% Define 1st Submatrix 
    A_1 = zeros(Nx,Nx);
    for ii = 2:(Nx-1)/2          
        A_1(ii,ii)   = -2*c.^2/dx.^2;
        A_1(ii,ii+1) =    c.^2/dx.^2;
        A_1(ii,ii-1) =    c.^2/dx.^2;
    end
    A_1(1,1) = -2*c.^2/dx.^2; 
    A_1(1,2) = 2*c.^2/dx.^2; 
    A_1((Nx-1)/2+1,(Nx-1)/2) =  c.^2/dx.^2; 
    A_1((Nx-1)/2+1,(Nx-1)/2+1)  = -c.^2/dx.^2;
  
    %% Define 1st Submatrix 
    A_2 = zeros(Nx,Nx);
    for ii = (Nx+1)/2:Nx-1           
        A_2(ii,ii)   = -2*c.^2/dx.^2;
        A_2(ii,ii+1) =    c.^2/dx.^2;
        A_2(ii,ii-1) =    c.^2/dx.^2;
    end
    A_2((Nx+1)/2-1,(Nx+1)/2-1) = -c.^2/dx.^2;
    A_2((Nx+1)/2-1,(Nx+1)/2) = c.^2/dx.^2; 
    A_2(Nx,Nx-1) =  2*c.^2/dx.^2; 
    A_2(Nx,Nx)   = -2*c.^2/dx.^2;

    %% Define Matrix in which we store A_h
    A_tmp = zeros(Nt,2*Nx,2*Nx); 
    for t = 1:Nt
        A_tmp(t,1:Nx,Nx+1:end) = eye(Nx, Nx);
    end
    
    %% Expand Matrix A to match Dimensions with A_tmp
    A = A_tmp;
    for ii = 1:Nt
        A(ii,Nx+1:end,1:Nx) = A_0(:,:); 
    end

    %% Define the Total Solution Matrix (first half: deflection, second half: d/dt of deflection)
    y_tot = zeros(2*Nx,N_tot); 
    u_tot = zeros(1,N_tot-1); 

    %% Set the Initial Condition
    for i=1:Nx
        y_tot(i,1) = 3*sin((dx*i*2*pi)/L);
    end
    
    %% Gradient Descent for the Optimal Control 

    kk = 1;
    limit_low_y = 1;
    limit_low_u = 1;
    limit_high_y = kk*N_MPC+1;
 
    while limit_high_y < N_tot 
        
        n = 0;
        u_MPC = zeros(1,Nt-1); 
        y_MPC = zeros(2*Nx,Nt);

        y_MPC(:,1) = y_tot(:,limit_low_y);
    
        [Ah] = random_Ah(A_1, A_2, Nt, K, Nx); % Get Randomized Matrix
        A_tmp(:,Nx+1:end,1:Nx) = Ah(:,:,:);
        
        crit = 1;
        
        y_MPC = state(u_MPC,A_tmp,B,y_MPC(:,1),dt,Nx,Nt);  
                                                          
        J = get_J(y_MPC, u_MPC,Q,dt,Nt);

        while (crit >= 1e-5) 
                
            phi = costate(y_MPC,A_tmp,Q,dt,Nx,Nt);
            grad_J = dt*B.'*phi + dt*R*u_MPC; % gradient
            y_nabla = state(grad_J,A_tmp,B,0*y_MPC(:,1),dt,Nx,Nt); % change in state resulting from gradient
            H = get_H(grad_J,y_nabla,Q,Nt,dt); % H: quadratic coefficient
            G = grad_J*grad_J.'; 
            alpha = G/H; % optimal stepsize 

            u_MPC = u_MPC - alpha*grad_J; % improve control
            y_MPC = state(u_MPC,A_tmp,B,y_MPC(:,1),dt,Nx,Nt); 
            Jnew = get_J(y_MPC, u_MPC,Q,dt,Nt); % update cost functional
        
            n = n + 1;
            crit = abs(J - Jnew) / J; 
            J = Jnew;    
        end

        fprintf("Step No. %d done!\n", kk);
        
        y_MPC = state(u_MPC,A,B,y_MPC(:,1),dt,Nx,Nt); % compute corresponding state

        limit_high_y = kk*N_MPC+1;
        limit_high_u = kk*N_MPC;

        y_tot(:,limit_low_y:limit_high_y) = y_MPC(:,1:N_MPC+1); % transfer y_MPC to total y_tot  
        u_tot(:,limit_low_u:limit_high_u) = u_MPC(:,1:N_MPC); % transfer u_MPC to total y_tot  

        limit_low_y = limit_high_y;
        limit_low_u = limit_high_u + 1;
         
        kk = kk + 1;

    end

    y_tot = y_tot(:,1:N_tot); % cut if neeeded
    u_tot = u_tot(:,1:N_tot-1); % cut if neeeded

end

function y = state(u,A,B,y0,dt,Nx,Nt) 
    y = zeros(2*Nx,Nt);
    y(:,1) = y0;
    I = eye(2*Nx,2*Nx);
    for tt = 2:Nt
        A_tmp = reshape(A(tt-1,:,:),2*Nx,2*Nx);
        y(:,tt) = (I - dt/2*A_tmp)\((I+dt/2*A_tmp)*y(:,tt-1) + dt*B*u(:,tt-1));
    end
end

function phi = costate(y,A,Q,dt,Nx,Nt) 
    phi = zeros(2*Nx,Nt-1);
    I = eye(2*Nx,2*Nx);
    A_tmp = A(end-1,:,:);
    A_tmp = reshape(A_tmp,2*Nx,2*Nx);
    phi(:,end) = (I-dt/2*A_tmp.')\(dt*Q*y(:,end)/2);
    for t = Nt-2:-1:1
        A_tmp = A(t,:,:);
        A_tmp = reshape(A_tmp,2*Nx,2*Nx);
        A_tmp2 = A(t+1,:,:); 
        A_tmp2 = reshape(A_tmp2,2*Nx,2*Nx);
        phi(:,t) = (I-dt/2*A_tmp.')\((I+dt/2*A_tmp2.')*phi(:,t+1)+ dt*Q*y(:,t+1));
    end
end

function J = get_J(X,u,Q,dt,NT)
    J = 0;
    for ii = 1:NT-1
        J = J + dt*X(:,ii).'*Q*X(:,ii)/4 + dt*X(:,ii+1).'*Q*X(:,ii+1)/4 + dt*u(ii).'*u(ii)/2;
    end
end 

function H = get_H(u,Xnabla,Q,NT,dt)
    H = 0;
    for ii = 1:NT-1
        H = H + dt*Xnabla(:,ii).'*Q*Xnabla(:,ii)/2 + dt*Xnabla(:,ii+1).'*Q*Xnabla(:,ii+1)/2 + dt*u(ii).'*u(ii);
    end
end


function Ah = random_Ah(A_1, A_2, Nt, K, Nx) % Returns Array of Matrices

    Ah = zeros(Nt,Nx,Nx); 
    h = Nt/K; 
   
    for j = 1:h:Nt
        r = randi([0 1],1,1); % Decide randomly, which of the matrices is found in the subset
                              % 0: upper right part (A_1), 1: lower right part (A_2)
        if r == 0
            for i = 1:h
                Ah(j+(i-1),:,:) = 2*A_1;
            end
        elseif r == 1
            for i = 1:h
                Ah(j+(i-1),:,:) = 2*A_2;
            end
        end
    end
end